﻿/*
 *
 *
 *    GenericMotor.cs
 *
 *    AutoCap - A generic motor interface.
 *
 *    License: GNU General Public License Version 3.0.
 *    
 *    Copyright (C) 2014 by Matthew K. Roberts, KK5JY. All rights reserved.
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *    
 *    This program is distributed in the hope that it will be useful, but
 *    WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    General Public License for more details.
 *    
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see: http://www.gnu.org/licenses/
 *    
 *
 */

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Net;

namespace KK5JY.AutoCap {
	/// <summary>
	/// A generic motor interface.
	/// </summary>
	/// <remarks>
	/// The generic key-value pairs used for each command message:
	///     Action=Pulse      - generates a pulse command, requires Speed and Duration also be sent
	///     Action=Move       - generates a move command, requires Speed also be sent
	///     Action=Brake      - changes the brake setting in the controller (requires 'Enable')
	///     Speed=n           - indicates the effort and direction, in the range [ -1.0m, 1.0m ]
	///     Duration=n        - indicates the pulse width, measured in seconds
	///     Enable=True|False - enable or disable the specific feature
	/// </remarks>
	public class GenericMotor : GenericConnector, ITimedMotor, IBrakedMotor, IReportsVersion {
		#region Public Properties
		/// <summary>
		/// Returns the version string.
		/// </summary>
		public string VersionString {
			get { return String.Format("Generic Motor - {0}:{1}", EndPoint.Address, EndPoint.Port); }
		}

		/// <summary>
		/// Indicates that all motor efforts should be reversed.
		/// </summary>
		public bool MotorReversed { get; set; }

		/// <summary>
		/// Returns the brake status set by the controller.
		/// </summary>
		public bool Brake { get; private set; }
		#endregion

		#region Public Methods
		/// <summary>
		/// Construct a new motor object.
		/// </summary>
		/// <param name="ep">The remote endpoint.</param>
		public GenericMotor(IPEndPoint ep) : base(ep) {
			// nop
		}

		/// <summary>
		/// Generate a movement pulse.
		/// </summary>
		/// <param name="speed">The normalized effort (negative is reversed).</param>
		/// <param name="duration">The duration, in seconds.</param>
		public bool Pulse(decimal speed, decimal duration) {
			if (speed > 1.0m) speed = 1.0m;
			if (speed < -1.0m) speed = -1.0m;
			if (MotorReversed) speed *= -1.0m;
			GenericMessage msg = new GenericMessage();
			msg.Add("Action", "Pulse");
			msg.Add("Speed", speed.ToString());
			msg.Add("Duration", duration.ToString());
			try {
				SendMessage(msg);
				return true;
			} catch {
				return false;
			}
		}

		/// <summary>
		/// Move or stop the motor.
		/// </summary>
		/// <param name="speed">The normalized speed, including direction sign.</param>
		public bool Move(decimal speed = 0) {
			if (speed > 1.0m) speed = 1.0m;
			if (speed < -1.0m) speed = -1.0m;
			if (MotorReversed) speed *= -1.0m;
			GenericMessage msg = new GenericMessage();
			msg.Add("Action", "Move");
			msg.Add("Speed", speed.ToString());
			try {
				SendMessage(msg);
				return true;
			} catch {
				return false;
			}
		}

		/// <summary>
		/// Enable or disable the brake.
		/// </summary>
		/// <param name="enabled">True to use braking, false otherwise.</param>
		public bool SetBraking(bool enabled) {
			GenericMessage msg = new GenericMessage();
			msg.Add("Action", "Brake");
			msg.Add("Enable", enabled.ToString());
			Brake = enabled; // TODO: not all controllers may support this
			try {
				SendMessage(msg);
				return true;
			} catch {
				return false;
			}
		}
		#endregion

		#region Event Handlers
		protected override void OnMalformedMessage(GenericMessageError e) {
			// nop
		}

		protected override void OnMessageReceived(GenericMessageEvent e) {
			// nop
		}
		#endregion
	}
}
