﻿/*
 *
 *
 *    LP100A.cs
 *
 *    Wattmeter interface.
 *
 *    License: GNU General Public License Version 3.0.
 *    
 *    Copyright (C) 2014 by Matthew K. Roberts, KK5JY. All rights reserved.
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *    
 *    This program is distributed in the hope that it will be useful, but
 *    WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    General Public License for more details.
 *    
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see: http://www.gnu.org/licenses/
 *    
 *
 */

// define this symbol to enable boxcar averaging
//#define BOXCAR

using System;
using System.Threading;
using System.IO.Ports;
using System.Net;

using KK5JY.IO;

namespace KK5JY.AutoCap {
	/// <summary>
	/// An LP-100A SWR meter.
	/// </summary>
	public class LP100A : ISensor, IStart, IReportsVersion, IDisposable {
		#region Constants
		/// <summary>
		/// The line ending for messages.
		/// </summary>
		private const char SOH = ';';

		/// <summary>
		/// The length of one data frame.
		/// </summary>
		private const int PacketLength = 43; // bytes
#if BOXCAR
		/// <summary>
		/// The length of the SWR boxcar average
		/// </summary>
		private const int BoxcarLength = 3; // samples
#endif
		#endregion

		#region Private Fields
		private Thread m_Thread;
		private IDeviceHelper m_Peer;
		private DateTimeOffset m_LastUpdate;
		private DateTimeOffset m_LastPoll;
		private string m_RxBuffer;
		private bool m_AutoPoll;
		private bool m_Quit;
#if BOXCAR
		private decimal[] m_Boxcar;
		private int m_BoxcarIndex;
#endif
		#endregion

		#region Public Events
		/// <summary>
		/// Raised when a new reading is taken.
		/// </summary>
		public event EventHandler<SensorDataEvent> SensorData;
		#endregion

		#region Public Properties
		/// <summary>
		/// Enable automatic polling.
		/// </summary>
		public bool AutoPoll {
			get {
				return m_AutoPoll;
			}
			set {
				if (m_AutoPoll != value) {
					m_AutoPoll = value;
					if (m_AutoPoll) {
						Start();
					} else {
						Stop();
					}
				}
			}
		}

		/// <summary>
		/// Indicates that the sensor should only try to read SWR (not power, reactance, etc.)
		/// </summary>
		public bool ReadOnlySWR { get; set; }

		/// <summary>
		/// Maximum time to wait for a Poll() response before re-polling.
		/// Only applies to auto-polling mode.
		/// </summary>
		private TimeSpan MaximumStale { get; set; }

		/// <summary>
		/// Minimum wait time between polls.  Gives the meter a chance to
		/// reset between polls.
		/// </summary>
		public TimeSpan MinimumPollInterval { get; set; }

		/// <summary>
		/// Return the version string.
		/// </summary>
		public string VersionString { get { return "LP-100A Serial Meter"; } }
		#endregion

		#region Public Methods
		private LP100A() {
			MaximumStale = new TimeSpan(0, 0, 0, 0, 250);
			MinimumPollInterval = new TimeSpan(0, 0, 0, 0, 50);

			m_RxBuffer = "";
#if BOXCAR
			m_Boxcar = new decimal[BoxcarLength];
			m_BoxcarIndex = 0;
#endif
		}

		/// <summary>
		/// Constructs a new object.
		/// </summary>
		/// <param name="portName">The serial port.</param>
		/// <param name="speed">The port speed (bps).</param>
		public LP100A(string portName, int speed = 115200) : this() {
			var port = new SerialPort(portName);
			m_Peer = new SerialPortHelper(port);
			m_Peer.DataReceived += DataReceived;
			port.BaudRate = speed;
			port.Parity = Parity.None;
			port.DataBits = 8;
			port.StopBits = StopBits.One;
		}

		/// <summary>
		/// Constructs a new object.
		/// </summary>
		/// <param name="addr">The remote address.</param>
		public LP100A(IPEndPoint addr) : this() {
			m_Peer = new SocketHelper(addr);
			m_Peer.DataReceived += DataReceived;
		}

		/// <summary>
		/// Manually request a new sample.
		/// </summary>
		public void Poll() {
			try {
				m_LastPoll = DateTimeOffset.UtcNow;
				m_Peer.Write("P");
			} catch {
				// nop
			}
		}

		/// <summary>
		/// Start autopolling.
		/// </summary>
		public void Start() {
			if (m_Thread != null) return;
			m_Quit = false;
			m_Peer.Start();
			m_Thread = new Thread(Worker) { Name = "LP100 AutoPoll Thread" };
			m_Thread.Start();
		}

		/// <summary>
		/// Stop autopolling.
		/// </summary>
		public void Stop() {
			m_Quit = true;
			try { m_Peer.Dispose(); } catch { /* nop */ }
			if (m_Thread == null) return;
			m_Thread.Join();
			m_Thread = null;
		}

		/// <summary>
		/// Stop autopolling.
		/// </summary>
		public void Dispose() {
			Stop();
		}
		#endregion

		#region Event Handlers and Utilities
		/// <summary>
		/// Raised when new serial data is received.
		/// </summary>
		private void DataReceived(object sender, ReceivedDataEvent e) {
			// read data and update the RX buffer
			m_RxBuffer += e.Data;

			// trim leading stuff; take only the most recent reading in the buffer
			for (int i = 1; i != m_RxBuffer.Length; ++i) {
				if (m_RxBuffer[i] == SOH) {
					m_RxBuffer = m_RxBuffer.Substring(i);
					break;
				}
			}

			// process whole messages
			while (m_RxBuffer.Length >= PacketLength) {
				var frame = m_RxBuffer.Substring(0, PacketLength);
				m_RxBuffer = m_RxBuffer.Substring(PacketLength);
				ProcessMessage(frame);
			}

			// if the buffer is now empty, poll again
			if (m_RxBuffer.Length == 0) {
				// wait to avoid hammering the meter
				if (!MinimumPollInterval.Equals(TimeSpan.Zero)) {
					var howLong = DateTimeOffset.UtcNow - m_LastPoll;
					if (howLong < MinimumPollInterval) {
						Thread.Sleep(MinimumPollInterval - howLong);
					}
				}

				// ask for the next reading
				Poll();
			}
		}

		/// <summary>
		/// Parse one complete record.
		/// </summary>
		private void ProcessMessage(string msg) {
			// From the LP-100 Manual:
			//     A   Increments Alarm Set Point selection 
			//     M   Increments Mode selection 
			//     F   Toggles Power Peak/Avg/Tune selection 
			//     P   Poll for data
			//
			// Update Format:
			//     Power, Z, Phase, SWR Alarm Set Point (0=off, 1=1.5, 2=2.0, 3=2.5, 4=3.0),
			//	       Callsign (6 digits with space padding), Power range: 0=High, 1=Mid, 2=Low, 
			//         Peak Hold Mode: 0=Average, 1= Peak Hold, dBm, SWR 

			// stamp the update
			m_LastUpdate = DateTimeOffset.UtcNow;

			// if there are no tied events, don't bother
			if (SensorData == null) {
				return;
			}

			// if there is no message data, bail
			if (msg.Length != PacketLength) return;
			msg = msg.Substring(1);

			// parse the message components
			var parts = msg.Split(',');
			if (parts.Length < 9) return;

			// DEBUG:
			//System.Diagnostics.Debug.WriteLine("DEBUG: LP100 data: " + msg);
#if BOXCAR
			var swr = 0m;
#endif
			var evData = new SensorData();
			try {
				// read non-SWR fields
				var power = Decimal.Parse(parts[0]);
				var z = Decimal.Parse(parts[1]);
				var phi = Decimal.Parse(parts[2]);

				if (ReadOnlySWR) {
					// setting ReadOnlySWR on an LP-100 is a really dumb idea, but
					//    we respect the flag, just the same
					evData.Power = null;
					evData.Z = null;
					evData.Phi = null;
				} else {
					evData.Power = power;
					evData.Z = z;
					evData.Phi = phi;
				}

				// read the SWR claimed by the meter, which is usually correct
				var swr = Decimal.Parse(parts[8]);

				// compute the SWR based on Z and angle, which is also usually correct,
				//   but is correct at different times than the claimed SWR
				var computedSWR = Convert.ToDecimal(
					SWR.ComputeSWR(
						50,
						Convert.ToDouble(z),
						Convert.ToDouble(phi)));

				// compute the difference between the two readings to detect errors
				var discrepency = Math.Abs(swr - computedSWR);

				//System.Diagnostics.Debug.WriteLine("DEBUG: LP-100 SWR = {0:0.00}, C_SWR = {1:0.00} Power = {2:0.0}", swr, computedSWR, power);

				// The LP-100 sometimes reports inconsistent readings when carrier
				//    is removed, possibly due to different numbers of samples being
				//    averaged for each reading.  Try to detect this window using
				//    the values from above, and fix the computed reading to compensate.
				if ((swr == 1.0m) && (discrepency >= 0.1m)) {
					computedSWR = 1.0m;
					evData.Power = 0.0m;
				}
#if BOXCAR
				swr = Decimal.Parse(parts[8]);
#else
				if (power == 0.0m) {
					evData.SWR = 1.0m;
				} else {
					// the 'swr' field from the meter freezes sometimes, so use the computed value
					evData.SWR = computedSWR;
				}
#endif
			} catch {
				return; // these are critical, failure not an option
			}
			try {
				evData.ID = parts[3].Trim();
			} catch {
				// nop - these are optional, just skip them if they fail to parse
			}
#if BOXCAR
			var avg = 0m;
			m_Boxcar[m_BoxcarIndex++] = swr;
			if (m_BoxcarIndex >= BoxcarLength) m_BoxcarIndex = 0;
			for (int i = 0; i != BoxcarLength; ++i) {
				avg += m_Boxcar[i];
			}
			avg /= BoxcarLength;
			m_Data.SWR = avg;
#endif
			// raise a new event
			var newReport = new SensorDataEvent { Data = evData };
			SensorData(this, newReport);
		}

		/// <summary>
		/// Worker thread.
		/// </summary>
		private void Worker() {
			// prime the pump
			Poll();

			// then just watchdog the link
			while (!m_Quit) {
				DateTimeOffset now = DateTimeOffset.UtcNow;

				// if the last reading is stale, ask for a new one
				if (now - m_LastUpdate > MaximumStale) {
					m_LastUpdate = now;
					Poll();
				}
				Thread.Sleep(MaximumStale);
			}
		}
		#endregion
	}
}
