﻿/*
 *
 *
 *    ProgramCLI.cs
 *
 *    AutoCap - The entry point for the CLI version.
 *
 *    License: GNU General Public License Version 3.0.
 *    
 *    Copyright (C) 2014 by Matthew K. Roberts, KK5JY. All rights reserved.
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *    
 *    This program is distributed in the hope that it will be useful, but
 *    WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    General Public License for more details.
 *    
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see: http://www.gnu.org/licenses/
 *    
 *
 */

using System;
using System.Xml;

namespace KK5JY.AutoCap {
	static class Program {
		// the parse states
		private enum ParseState {
			// not parsing anything, ignore characters until frame properly started
			None,
			// frame started with '#', parse command
			Command,
			// command parsing finished, but '=' found, so parse the argument
			Arg,
			// frame terminated with ';', so process entire command/argument pair
			Done,
		}

		// the core object
		private static Core m_Core;

		// user pulse settings
		private static decimal m_UserPulse, m_UserEffort;
		
		// flag for console echo
		private static bool m_Echo;

		/// <summary>
		/// The main entry point for the application.
		/// </summary>
		[STAThread]
		public static void Main(string [] args) {
			StringComparison cmp = StringComparison.InvariantCultureIgnoreCase;
			bool autoStart = false;

			foreach (var nextArg in args) {
				if (String.Equals(nextArg, "/s", cmp) || String.Equals(nextArg, "-s", cmp)) {
					autoStart = true;
				}
			}

			Console.WriteLine("#INFO={0};", Platform.VersionString);
			#if DEBUG
			Console.WriteLine("#DEBUG=config path -> {0};", Platform.ConfigFilePath);
			Console.WriteLine("#DEBUG=profile path -> {0};", Platform.UserProfileFolder);
			#endif
			Platform.CreateProfileFolders();

			// read the configuration
			m_Core = new Core();
			if (!System.IO.File.Exists(Platform.ConfigFilePath)) {
				m_Core.Config.SensorType = SensorTypes.AutoCapV1;
				m_Core.Config.MotorType = MotorTypes.AutoCapV1;
				var config = m_Core.Config.Serialize();
				config.Save(Platform.ConfigFilePath);
				Console.WriteLine("#WARNING=Using default configuration;");
			} else {
				XmlDocument doc = new XmlDocument();
				doc.Load(Platform.ConfigFilePath);
				if (doc.DocumentElement.HasAttribute("UserPulse")) {
					m_UserPulse = Decimal.Parse(doc.DocumentElement.GetAttribute("UserPulse"));
				} else {
					m_UserPulse = 0.100m;
				}
				if (doc.DocumentElement.HasAttribute("UserEffort")) {
					m_UserEffort = Decimal.Parse(doc.DocumentElement.GetAttribute("UserEffort"));
				} else {
					m_UserEffort = 0.100m;
				}
				m_Core.Config.Deserialize(doc);
			}

			DateTimeOffset lastUpdate = DateTimeOffset.UtcNow;
			TimeSpan autoValue = TimeSpan.FromSeconds(0.0);
			ParseState state = ParseState.None;
			string cmd = "", arg = "";
			bool doUpdate = false;
			bool running = false;

			// queue up a command if auto-starting
			if (autoStart) {
				cmd = "START";
				arg = "";
				state = ParseState.Done;
			}

			// forever
			while (true) {
				while (Console.KeyAvailable) {
					char ch = (char)Console.ReadKey(true).KeyChar;
					if (m_Echo && (state != ParseState.None)) {
						Console.Write(ch);
						Console.Out.Flush();
					}
					if (ch == '#') {
						cmd = "";
						arg = "";
						state = ParseState.Command;
						if (m_Echo) {
							Console.Write(ch);
							Console.Out.Flush();
						}
					} else if (ch == '=') {
						if (String.IsNullOrEmpty(cmd)) state = ParseState.None;
						else state = ParseState.Arg;
					} else if (ch == ';') {
						if (String.IsNullOrEmpty(cmd)) state = ParseState.None;
						else state = ParseState.Done;
						if (m_Echo) Console.WriteLine();
						break;
					} else {
						if (state == ParseState.Command) {
							cmd += ch;
						}
						if (state == ParseState.Arg) {
							arg += ch;
						}
					}
				}

                #region Process Command
                if (state == ParseState.Done) {
					cmd = cmd.Trim();
					arg = arg.Trim();

					var ok = false;
					var quit = false;

					if (String.Equals(cmd, "START", cmp)) { // start the tuner
						// start the core
						try {
							m_Core.Start();
							running = true;
							ok = true;
						} catch (ApplicationException ex) {
							Console.WriteLine("#ERRINFO=Could not start core: {0};", ex.Message);
							if (autoStart) return;
						} catch (Exception ex) {
							#if DEBUG
							Console.WriteLine("#ERRINFO=Could not start core: {0}: {1}: {2};", ex.GetType().Name, ex.Message, ex.StackTrace);
							#else
							Console.WriteLine("#ERRINFO=Could not start core: {0}: {1};", ex.GetType().Name, ex.Message);
							#endif
							if (autoStart) return;
						}
					} else if (String.Equals(cmd, "STOP", cmp)) { // stop the tuner
						Stop();
						running = false;
						autoValue = TimeSpan.Zero;
						ok = true;
					} else if (String.Equals(cmd, "EXIT", cmp) || String.Equals(cmd, "QUIT", cmp)) {
						// stop the server
						Stop();
						running = false;
						autoValue = TimeSpan.Zero;
						ok = true;
						quit = true;
					} else if (String.Equals(cmd, "ECHO", cmp)) {
                        if (String.IsNullOrWhiteSpace(arg)) {
                            arg = m_Echo.ToString();
                            ok = true;
                        } else {
                            bool newValue;
                            if (Boolean.TryParse(arg, out newValue)) {
                                m_Echo = newValue;
                                ok = true;
                            }
                        }
					} else if (String.Equals(cmd, "ALL", cmp)) { // single display of runtime parameters
						if (running) {
							doUpdate = true;
							ok = true;
						}
					} else if (String.Equals(cmd, "AUTO", cmp)) { // auto-display of runtime parames
						if (String.IsNullOrWhiteSpace(arg)) {
							arg = autoValue.TotalSeconds.ToString();
							ok = true;
						} else {
							double newVal;
							if (Double.TryParse(arg, out newVal)) {
								if (newVal == 0.0) {
									autoValue = TimeSpan.Zero;
									ok = true;
								} else if (newVal > 0.100) {
									autoValue = TimeSpan.FromSeconds(newVal);
									ok = true;
								}
							}
						}
					} else if (String.Equals(cmd, "PAUSE", cmp)) {
						if (String.IsNullOrWhiteSpace(arg)) {
							arg = m_Core.Automatic.ToString();
							ok = true;
						} else {
							bool newVal;
							if (Boolean.TryParse(arg, out newVal)) {
								m_Core.Automatic = newVal;
							}
						}
					} else if (String.Equals(cmd, "SAVE", cmp)) {
						var path = Platform.ConfigFilePath;
						if (!String.IsNullOrWhiteSpace(arg)) {
							path = arg.Trim();
						}
						if (!(path.StartsWith("/") || path.StartsWith("."))) {
							path = System.IO.Path.Combine(Platform.UserProfileFolder, path);
						}
						if (String.IsNullOrEmpty(System.IO.Path.GetExtension(path))) {
							path += ".xml";
						}

						try {
							var doc = m_Core.Config.Serialize();
							doc.DocumentElement.SetAttribute("UserPulse", m_UserPulse.ToString());
							doc.DocumentElement.SetAttribute("UserEffort", m_UserEffort.ToString());
							doc.DocumentElement.SetAttribute("FreqHint", m_Core.FrequencyHint.ToString());
							doc.Save(path);
							ok = true;
						} catch (Exception ex) {
							Console.WriteLine("#ERRINFO=Could not save profile: {0};", ex.Message);
						}
					} else if (String.Equals(cmd, "LOAD", cmp)) {
						var path = Platform.ConfigFilePath;
						if (!String.IsNullOrWhiteSpace(arg)) {
							path = arg.Trim();
						}
						if (!(path.StartsWith("/") || path.StartsWith("."))) {
							path = System.IO.Path.Combine(Platform.UserProfileFolder, path);
						}
						if (String.IsNullOrEmpty(System.IO.Path.GetExtension(path))) {
							path += ".xml";
						}

						try {
							var doc = new XmlDocument();
							if (String.IsNullOrEmpty(path)) {
								path = Platform.ConfigFilePath;
							}

							doc.Load(path);
							m_Core.Config.Deserialize(doc);

							// consume custom attributes from the root element
							if (doc.DocumentElement.HasAttribute("UserPulse")) {
								m_UserPulse = Decimal.Parse(doc.DocumentElement.GetAttribute("UserPulse"));
							}
							if (doc.DocumentElement.HasAttribute("UserEffort")) {
								m_UserEffort = Decimal.Parse(doc.DocumentElement.GetAttribute("UserEffort"));
							}
							if (doc.DocumentElement.HasAttribute("FreqHint")) {
								m_Core.FrequencyHint = Int32.Parse(doc.DocumentElement.GetAttribute("FreqHint"));
							}
							ok = true;
						} catch (Exception ex) {
							Console.WriteLine("#ERRINFO=Could not load profile: {0};", ex.Message);
						}
					} else if (String.Equals(cmd, "USERPULSE", cmp)) {
						if (String.IsNullOrWhiteSpace(arg)) {
							arg = m_UserPulse.ToString();
							ok = true;
						} else {
							decimal newValue;
							if (Decimal.TryParse(arg, out newValue)) {
								m_UserPulse = newValue;
								ok = true;
							}
						}
					} else if (String.Equals(cmd, "USEREFFORT", cmp)) {
						if (String.IsNullOrWhiteSpace(arg)) {
							arg = m_UserPulse.ToString();
							ok = true;
						} else {
							decimal newValue;
							if (Decimal.TryParse(arg, out newValue)) {
								m_UserEffort = newValue;
								ok = true;
							}
						}
					} else if (String.Equals(cmd, "PULSEUP", cmp)) {
						m_Core.Pulse(m_UserEffort, m_UserPulse);
                        ok = true;
					} else if (String.Equals(cmd, "PULSEDOWN", cmp)) {
						m_Core.Pulse(m_UserEffort, -m_UserPulse);
                        ok = true;
					} else if (String.Equals(cmd, "THRESHOLD", cmp)) {
                        var parts = arg.Split(',');
                        Int32 row;
                        if (Int32.TryParse(parts[0], out row) && row >= 0 && row < 4) {
                            if (parts.Length == 1) {        // get
                                arg = String.Format("{0},{1},{2},{3},{4}",
                                    row,
                                    m_Core.Config.Thresholds[row].SWR,
                                    m_Core.Config.Thresholds[row].Size,
                                    m_Core.Config.Thresholds[row].Effort,
									m_Core.Config.Thresholds[row].IsStep);
                                ok = true;
                            } else if (parts.Length == 5) { // set
                                decimal swr, size, effort;
								bool isStep;
                                if (Decimal.TryParse(parts[1], out swr) &&
                                    Decimal.TryParse(parts[2], out size) &&
                                    Decimal.TryParse(parts[3], out effort) &&
									Boolean.TryParse(parts[4], out isStep)) {
                                    m_Core.Config.Thresholds[row].SWR = swr;
                                    m_Core.Config.Thresholds[row].Size = size;
                                    m_Core.Config.Thresholds[row].Effort = effort;
									m_Core.Config.Thresholds[row].IsStep = isStep;
                                    arg = String.Format("{0},{1},{2},{3},{4}",
                                        row,
                                        m_Core.Config.Thresholds[row].SWR,
                                        m_Core.Config.Thresholds[row].Size,
                                        m_Core.Config.Thresholds[row].Effort,
										m_Core.Config.Thresholds[row].IsStep);
                                    ok = true;
                                }
                            }
                        }
                    } else if (String.Equals(cmd, "MOTORTYPE", cmp)) {
                        if (String.IsNullOrWhiteSpace(arg)) {
                            arg = m_Core.Config.MotorType.ToString();
                            ok = true;
                        } else {
                            MotorTypes newValue;
                            if (Enum.TryParse(arg, true, out newValue)) {
                                m_Core.Config.MotorType = newValue;
                                ok = true;
                            }
                        }
					} else if (String.Equals(cmd, "MOTORCONTROLPORT", cmp)) {
                        if (String.IsNullOrWhiteSpace(arg)) {
                            arg = m_Core.Config.MotorControlPort.ToString();
                            ok = true;
                        } else {
                            if (!String.IsNullOrEmpty(arg)) {
                                // TODO: validate the new port string
                                m_Core.Config.MotorControlPort = arg;
                                ok = true;
                            }
                        }
					} else if (String.Equals(cmd, "MOTOROUTPUTPORT", cmp)) {
                        if (String.IsNullOrWhiteSpace(arg)) {
                            arg = m_Core.Config.MotorOutputPort.ToString();
                            ok = true;
                        } else {
                            int newValue;
                            if (Int32.TryParse(arg, out newValue)) {
                                m_Core.Config.MotorOutputPort = newValue;
                                ok = true;
                            }
                        }
					} else if (String.Equals(cmd, "MOTORREVERSE", cmp)) {
                        if (String.IsNullOrWhiteSpace(arg)) {
                            arg = m_Core.Config.MotorPolarityReversed.ToString();
                            ok = true;
                        } else {
                            bool newValue;
                            if (Boolean.TryParse(arg, out newValue)) {
                                m_Core.Config.MotorPolarityReversed = newValue;
                                ok = true;
                            }
                        }
                    } else if (String.Equals(cmd, "MOTORRELAYREVERSE", cmp)) {
                        if (String.IsNullOrWhiteSpace(arg)) {
                            arg = m_Core.Config.MotorRelayRolesReversed.ToString();
                            ok = true;
                        } else {
                            bool newValue;
                            if (Boolean.TryParse(arg, out newValue)) {
                                m_Core.Config.MotorRelayRolesReversed = newValue;
                                ok = true;
                            }
                        }
                    } else if (String.Equals(cmd, "MOTORBRAKE", cmp)) {
                        if (String.IsNullOrWhiteSpace(arg)) {
                            arg = m_Core.Config.MotorBrakingEnabled.ToString();
                            ok = true;
                        } else {
                            bool newValue;
                            if (Boolean.TryParse(arg, out newValue)) {
                                m_Core.Config.MotorBrakingEnabled = newValue;
                                ok = true;
                            }
                        }
                    } else if (String.Equals(cmd, "SENSORTYPE", cmp)) {
                        if (String.IsNullOrWhiteSpace(arg)) {
                            arg = m_Core.Config.SensorType.ToString();
                            ok = true;
                        } else {
                            SensorTypes newValue;
                            if (Enum.TryParse(arg, true, out newValue)) {
                                m_Core.Config.SensorType = newValue;
                                ok = true;
                            }
                        }
                    } else if (String.Equals(cmd, "SENSORPORT", cmp)) {
                        if (String.IsNullOrWhiteSpace(arg)) {
                            arg = m_Core.Config.SensorPort.ToString();
                            ok = true;
                        } else {
                            if (!String.IsNullOrEmpty(arg)) {
                                // TODO: validate the new port string
                                m_Core.Config.SensorPort = arg;
                                ok = true;
                            }
                        }
					} else if (String.Equals(cmd, "READDELAY", cmp)) {
						if (String.IsNullOrWhiteSpace(arg)) {
							arg = m_Core.Config.ReadDelay.ToString();
							ok = true;
						} else {
							decimal newValue;
							if (Decimal.TryParse(arg, out newValue)) {
								m_Core.Config.ReadDelay = newValue;
								ok = true;
							}
						}
					} else if (String.Equals(cmd, "STEPPERRATE", cmp)) {
						if (String.IsNullOrWhiteSpace(arg)) {
							arg = m_Core.Config.MaxStepRate.ToString();
							ok = true;
						} else {
							decimal newValue;
							if (Decimal.TryParse(arg, out newValue)) {
								m_Core.Config.MaxStepRate = Convert.ToUInt32(newValue);
								ok = true;
							}
						}
					} else if (String.Equals(cmd, "SENSORMINPOWER", cmp)) {
                        if (String.IsNullOrWhiteSpace(arg)) {
                            arg = m_Core.Config.MinPower.ToString();
                            ok = true;
                        } else {
                            decimal newValue;
                            if (Decimal.TryParse(arg, out newValue)) {
                                m_Core.Config.MinPower = newValue;
                                ok = true;
                            }
                        }
                    } else if (String.Equals(cmd, "SENSOREPSILON", cmp)) {
                        if (String.IsNullOrWhiteSpace(arg)) {
                            arg = m_Core.Config.SensorEpsilon.ToString();
                            ok = true;
                        } else {
                            decimal newValue;
                            if (Decimal.TryParse(arg, out newValue)) {
                                m_Core.Config.SensorEpsilon = newValue;
                                ok = true;
                            }
                        }
                    } else if (String.Equals(cmd, "RADIOTYPE", cmp)) {
                        if (String.IsNullOrWhiteSpace(arg)) {
                            arg = m_Core.Config.RadioType.ToString();
                            ok = true;
                        } else {
                            RadioTypes newValue;
                            if (Enum.TryParse(arg, true, out newValue)) {
                                m_Core.Config.RadioType = newValue;
                                ok = true;
                            }
                        }
                    } else if (String.Equals(cmd, "RADIOADDRESS", cmp)) {
                        if (String.IsNullOrWhiteSpace(arg)) {
                            arg = m_Core.Config.RadioHost.ToString();
                            ok = true;
                        } else {
                            m_Core.Config.RadioHost = arg;
                            ok = true;
                        }
                    } else if (String.Equals(cmd, "RADIOPORT", cmp)) {
                        if (String.IsNullOrWhiteSpace(arg)) {
                            arg = m_Core.Config.RadioPort.ToString();
                            ok = true;
                        } else {
                            int newValue;
                            if (Int32.TryParse(arg, out newValue)) {
                                m_Core.Config.RadioPort = newValue;
                                ok = true;
                            }
                        }
                    }

					var resp = (ok ? "#OK:" : "#ERR:") + cmd;
					if (!String.IsNullOrEmpty(arg)) {
						resp = resp + '=' + arg;
					}
					resp += ';';
					Console.Write(resp);
					if (m_Echo) Console.WriteLine();
					Console.Out.Flush();

					state = ParseState.None;
					cmd = "";
					arg = "";

					if (quit) {
						// next prompt on newline
						Console.WriteLine();
						return;
					}
                }
                #endregion

                // TODO: auto-update the 'all' display
				var now = DateTimeOffset.UtcNow;
				if ((!autoValue.Equals(TimeSpan.Zero)) && ((now - lastUpdate) >= autoValue)) {
					lastUpdate = now;
					doUpdate = true;
				}

				// wait for the user to cancel the algorithm
				if (doUpdate) {
					doUpdate = false;
					var data = m_Core.Data;
					if (data != null) {
						System.Console.WriteLine(
							"#ALL={0},{1},{2},{3};",
							m_Core.Data.SWR, m_Core.Data.Power, m_Core.MotorEffort, m_Core.Frequency
						);
					}
				}

				// be nice
				System.Threading.Thread.Sleep(100);
			}
		}

		private static void Stop() {
			// stop the core
			try {
				m_Core.Stop();
			} catch {
				// nop
			}
		}
	}
}
